from typing import Literal, NamedTuple

from .http_clients import AbstractHTTPClient, HTTPXClient
from .models import (
    CompileStage,
    ExecutionOutput,
    File,
    Package,
    RunStage,
    Runtime,
)
from .piston import PistonClient

__all__ = (
    "AbstractHTTPClient",
    "HTTPXClient",
    "Runtime",
    "Package",
    "File",
    "RunStage",
    "CompileStage",
    "ExecutionOutput",
    "PistonClient",
)

__version__ = "1.0.1"
__title__ = "ultra-piston"
__author__ = "Krish Mohan M."
__license__ = "MIT"
__copyright__ = "Copyright 2025-present, Krish Mohan M."


class VersionInfo(NamedTuple):
    major: str
    minor: str
    patch: str
    releaselevel: Literal["alpha", "beta", "final"]


def _expand() -> VersionInfo:
    v = __version__.split(".")
    level_types = {"a": "alpha", "b": "beta"}
    level = level_types.get(v[-1], "final")
    return VersionInfo(major=v[0], minor=v[1], patch=v[2], releaselevel=level)  # pyright:ignore[reportArgumentType]


version_info: VersionInfo = _expand()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Tuple


__all__ = ("MISSING",)


class _MissingSentinel:
    __slots__: Tuple[str, ...] = ()

    def __eq__(self, other: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> str:
        return "MISSING"


MISSING: Any = _MissingSentinel()
"""Used in areas where an attribute doesn't have a value by default but
gets defined during runtime. Lesser type checking would be required by using
this, opposed to using some other default value such as ``None``."""



from __future__ import annotations

import asyncio
import functools
import time
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict

try:
    import httpx
except ModuleNotFoundError:
    pass

from .errors import (
    BadRequestError,
    InternalServerError,
    MissingDataError,
    NotFoundError,
    TooManyRequestsError,
    UnexpectedStatusError,
)
from .utils import MISSING

if TYPE_CHECKING:
    from asyncio import Queue
    from types import CoroutineType
    from typing import (
        Any,
        Awaitable,
        Callable,
        Dict,
        Optional,
        ParamSpec,
        TypeVar,
        Union,
    )

    from httpx import Response

    P = ParamSpec("P")
    R = TypeVar("R")


__all__ = ("AbstractHTTPClient", "HTTPXClient")


class AbstractHTTPClient(ABC):
    r"""
    An abstract base class for HTTP clients used to communicate with external services.

    Attributes
    ----------
    driver : Optional[str]
        | The name of the HTTP driver or implementation used.
    rate_limit : Optional[Union[int, float]]
        | Ratelimit to set for the dispatched requests.
        | Takes in a integer / float of the amount of delay (in seconds) between each request.
        | Defaults to 1 request per second.
    base_url : Optional[str]
        | The base URL for the API endpoint.
    headers : Optional[Dict[str, str]]
        | Default headers to include with every request.
    """

    def __init__(self) -> None:
        self.driver: str = MISSING
        self.rate_limit: Union[int, float] = MISSING
        self.base_url: str = MISSING
        self.headers: Dict[str, str] = MISSING

    def _get_rate_limit(self) -> Union[int, float]:
        r"""Retrieve the configured request rate limit.

        Returns
        -------
        The rate limit value.

        Raises
        ------
        :py:exc:`ultra_piston.errors.MissingDataError`
            | If no rate limit is set.
        """

        if self.rate_limit is MISSING:
            raise MissingDataError(
                f"Missing valid value for the attribute `self.rate_limit` of {self.__class__}."
            )
        return self.rate_limit

    def _get_base_url(self) -> str:
        r"""Retrieve the configured base URL for the API.

        Returns
        -------
        The base URL.

        Raises
        ------
        :py:exc:`ultra_piston.errors.MissingDataError`
            | If no base URL is set.
        """

        if self.base_url is MISSING:
            raise MissingDataError(
                f"Missing valid value for the attribute `self.base_url` of {self.__class__}."
            )
        return self.base_url

    def _get_headers(self) -> Dict[str, str]:
        r"""Retrieve the default HTTP headers for requests.

        Returns
        -------
        The request headers.

        Raises
        ------
        :py:exc:`ultra_piston.errors.MissingDataError`
            | If no headers are set.
        """

        if self.headers is MISSING:
            raise MissingDataError(
                f"Missing valid value for the attribute `self.headers` of {self.__class__}."
            )
        return self.headers

    def _get_http_payload(self, endpoint: str = "") -> Dict[str, Any]:
        r"""Construct the full HTTP payload with URL and headers.

        Parameters
        ----------
        endpoint
            | The endpoint path to append to the base URL.

        Returns
        -------
        Dictionary containing the full URL and headers.

        Raises
        ------
        :py:exc:`ultra_piston.errors.MissingDataError`
            | If no base URL or headers are set.
        """

        BASE_URL = self._get_base_url() + endpoint
        HEADERS = self._get_headers()

        return {"url": BASE_URL, "headers": HEADERS}

    @abstractmethod
    def get(self, endpoint: str) -> Any:
        r"""Send a synchronous GET request to the given endpoint.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.

        Returns
        -------
        The server response.
        """

    @abstractmethod
    async def get_async(self, endpoint: str) -> Any:
        r"""Send an asynchronous GET request to the given endpoint.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.

        Returns
        -------
        The server response.
        """

    @abstractmethod
    def post(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        r"""Send a synchronous POST request with optional JSON data.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.
        json_data
            | The JSON payload to send.

        Returns
        -------
        The server response.
        """

    @abstractmethod
    async def post_async(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        r"""Send an asynchronous POST request with optional JSON data.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.
        json_data
            | The JSON payload to send.

        Returns
        -------
        The server response.
        """

    @abstractmethod
    def delete(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        r"""Send a synchronous DELETE request with optional JSON data.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.
        json_data
            | The JSON payload to send.

        Returns
        -------
        The server response.
        """

    @abstractmethod
    async def delete_async(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        r"""Send an asynchronous DELETE request with optional JSON data.

        Parameters
        ----------
        endpoint
            | The API endpoint to request.
        json_data
            | The JSON payload to send.

        Returns
        -------
        The server response
        """


class HTTPXClient(AbstractHTTPClient):
    def __init__(self) -> None:
        super().__init__()

        self.driver: str = "httpx"
        self._last_request: Optional[datetime] = None
        self._request_queue: Queue[Awaitable[Any]] = asyncio.Queue(maxsize=1)

    def _validate_response_status(self, response: Response) -> Any:
        if 300 > response.status_code > 199:
            return response.json()

        elif response.status_code == 400:
            raise BadRequestError(str(response.url))

        elif response.status_code == 404:
            raise NotFoundError(str(response.url))

        elif response.status_code == 429:
            raise TooManyRequestsError(str(response.url))

        elif response.status_code == 500:
            raise InternalServerError(str(response.url))

        else:
            raise UnexpectedStatusError(
                response.status_code, str(response.url)
            )

    @staticmethod
    def sync_ratelimit(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            instance: HTTPXClient = args[0]  # pyright: ignore[reportAssignmentType]

            if instance._last_request:
                now = datetime.now()
                future_request_time = instance._last_request + timedelta(
                    seconds=instance._get_rate_limit()
                )
                if now < future_request_time:
                    cool_down = (future_request_time - now).seconds
                    time.sleep(cool_down)

            result = func(*args, **kwargs)

            instance._last_request = datetime.now()

            return result

        return wrapper

    @staticmethod
    def async_ratelimit(
        func: Callable[P, CoroutineType[Any, Any, R]],
    ) -> Callable[P, CoroutineType[Any, Any, R]]:
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            event_loop = asyncio.get_event_loop()
            instance: HTTPXClient = args[0]  # pyright: ignore[reportAssignmentType]

            if instance._last_request is not None:
                now = datetime.now()
                valid_future_stamp = instance._last_request + timedelta(
                    seconds=instance._get_rate_limit()
                )
                if now < valid_future_stamp:
                    # Rate limited
                    cool_down = (valid_future_stamp - now).seconds
                    await asyncio.sleep(cool_down)

            instance._last_request = datetime.now()

            future_await = func(*args, **kwargs)
            task_fetch = event_loop.create_task(
                instance._create_queue(future_await)
            )
            task_data = await asyncio.wait([task_fetch])
            return task_data[0].pop().result()

        return wrapper

    async def _create_queue(self, coro: Awaitable[R]) -> R:
        await self._request_queue.put(coro)
        awaitable = await self._request_queue.get()
        return await awaitable

    @sync_ratelimit
    def get(self, endpoint: str) -> Any:
        payload = self._get_http_payload(endpoint)
        response = httpx.get(**payload)
        return self._validate_response_status(response=response)

    @async_ratelimit
    async def get_async(self, endpoint: str) -> Any:
        payload = self._get_http_payload(endpoint)

        async with httpx.AsyncClient() as client:
            response = await client.get(**payload)

        return self._validate_response_status(response=response)

    @sync_ratelimit
    def post(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        payload = self._get_http_payload(endpoint)
        if json_data:
            payload["json"] = json_data

        response = httpx.post(**payload)
        return self._validate_response_status(response=response)

    @async_ratelimit
    async def post_async(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        payload = self._get_http_payload(endpoint)
        if json_data:
            payload["json"] = json_data

        async with httpx.AsyncClient() as client:
            response = await client.post(**payload)

        return self._validate_response_status(response=response)

    @sync_ratelimit
    def delete(
        self, endpoint: str, json_data: Dict[Any, Any] | None = None
    ) -> Any:
        payload = self._get_http_payload(endpoint)
        payload["method"] = "DELETE"
        if json_data:
            payload["json"] = json_data

        response = httpx.request(**payload)
        return self._validate_response_status(response=response)

    @async_ratelimit
    async def delete_async(
        self, endpoint: str, json_data: Optional[Dict[Any, Any]] = None
    ) -> Any:
        payload = self._get_http_payload(endpoint)
        payload["method"] = "DELETE"
        if json_data:
            payload["json"] = json_data

        async with httpx.AsyncClient() as client:
            response = await client.request(**payload)

        return self._validate_response_status(response=response)



from typing import List, Literal, Optional, Union

from pydantic import BaseModel

__all__ = (
    "Runtime",
    "Package",
    "File",
    "RunStage",
    "CompileStage",
    "ExecutionOutput",
)

BaseModel.__init__.__doc__ = None


class Runtime(BaseModel):
    r"""Represents a supported language runtime.

    Attributes
    ----------
    language : :class:`str`
        | The programming language.
    version : :class:`str`
        | The specific version of the language (e.g., "3.10.0").
    aliases : List[:class:`str`]
        | Alternate names or shortcuts for the language.
    runtime : :class:`Optional` [:class:`str`]
        | The runtime environment identifier, if applicable.
    """

    language: str
    version: str
    aliases: List[str]
    runtime: Optional[str] = None


class Package(BaseModel):
    r"""Represents a package or dependency available for a specific language.

    Attributes
    ----------
    language : :class:`str`
        | The programming language the package is for.
    language_version : :class:`str`
        | The version of the language the package is associated with.
    installed : :class:`bool`
        | Whether the package is installed and available.
    """

    language: str
    language_version: str
    installed: bool


class File(BaseModel):
    r"""Represents a file to be sent for execution.

    Attributes
    ----------
    name : :class:`Optional` [:class:`str`]
        | The name of the file (e.g., "main.py").
    content : :class:`str`
        | The raw contents of the file.
    encoding : :class:`Literal["base64", "hex", "utf8"]`
        | The encoding format used for the content. Defaults to "utf8".
    """

    name: Optional[str] = None
    content: str
    encoding: Literal["base64", "hex", "utf8"] = "utf8"


class RunStage(BaseModel):
    r"""Represents the result of the runtime execution stage.

    Attributes
    ----------
    code : :class:`Optional[:class:`int`]`
        | Exit code of the execution process.
    output : :class:`str`
        | Combined standard output and or error.
    stderr : :class:`str`
        | Standard error stream.
    stdout : :class:`str`
        | Standard output stream.
    signal : :class:`Optional` [:class:`str`]
        | Signal that caused the process to terminate, if any.
    """

    code: Optional[int] = None
    output: str
    stderr: str
    stdout: str
    signal: Optional[str] = None


class CompileStage(BaseModel):
    r"""Represents the result of the compilation stage, if applicable.

    Attributes
    ----------
    code : :class:`Optional` [:class:`int`]
        | Exit code of the compiler.
    output : :class:`str`
        | Combined standard output and error from the compiler.
    stderr : :class:`str`
        | Standard error stream from the compiler.
    stdout : :class:`str`
        | Standard output stream from the compiler.
    signal : :class:`Optional` [:class:`str`]
        | Signal that caused the compiler process to terminate, if any.
    """

    code: Optional[int] = None
    output: str
    stderr: str
    stdout: str
    signal: Optional[str] = None


class ExecutionOutput(BaseModel):
    r"""Represents the complete output from a code execution request.

    Attributes
    ----------
    language : :class:`str`
        | The language used to execute the code.
    version : :class:`str`
        | The language version used.
    run : :py:class:`RunStage`
        | Output from the runtime execution stage.
    compile : :class:`Optional` [:py:class:`CompileStage`]
        | Output from the compilation stage, if compilation was required.
    compile_memory_limit : :class:`Optional` [:class:`int`]
        | Memory limit (in bytes) used during compilation.
    compile_timeout : :class:`Optional` [Union[:class:`int`, :class:`float`]]
        | Timeout (in milliseconds) for the compilation stage.
    """

    language: str
    version: str
    run: RunStage
    compile: Optional[CompileStage] = None
    compile_memory_limit: Optional[int] = None
    compile_timeout: Optional[Union[int, float]] = None



from __future__ import annotations

import functools
import importlib
import logging
from typing import TYPE_CHECKING

import aiocache

from .http_clients import HTTPXClient
from .models import ExecutionOutput, File, Package, Runtime

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Type, Union

    from .http_clients import AbstractHTTPClient


__all__ = ("PistonClient",)
_logger = logging.getLogger(__name__)


class PistonClient:
    r"""
    The main client to interact with the piston-api.

    Parameters
    ----------
    api_key
        | The API key to use if any.
    rate_limit
        | Ratelimit to set for the dispatched requests.
        | Takes in a integer / float of the amount of delay between each request.
        | Defaults to 1 request per second.
    app_name
        | Name of your app / project. To be used by the HTTP client's User-Agent.
    base_url
        | Base URL of the API.
    http_client
        | The http client through which the requests are made with.
    http_client_kwargs
        | The attributes to set to the http client.
        | Useful when you want to make your own HTTP client.
    """

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        rate_limit: Union[float, int] = 1.0,
        app_name: str = "Ultra-Piston-Wrapper",
        base_url: str = "https://emkc.org/api/v2/piston/",
        http_client: Type[AbstractHTTPClient] = HTTPXClient,
        **http_client_kwargs: Any,
    ) -> None:
        self._http_client: AbstractHTTPClient = http_client()
        driver_version: Optional[str] = None

        if not self._http_client.driver:
            raise ValueError(
                f"No http `driver` was specified of `http_client={http_client}`. "
                f"Please make sure you have specified the value of the `driver` attribute in `{http_client}`. "
                "Example: httpx, aiohttp, requests, etc."
            )

        try:
            driver_lib = importlib.import_module(self._http_client.driver)

        except ModuleNotFoundError as error:
            raise ModuleNotFoundError(
                f"Couldn't find the specified HTTP driver: `{self._http_client.driver}` "
                f"from the passed `http_client={http_client}`. Please make sure you have installed "
                f"`{self._http_client.driver}` before running your project."
            ) from error

        driver_version = getattr(driver_lib, "__version__")
        if not driver_version:
            version_info = getattr(driver_lib, "version_info")
            if version_info:
                try:
                    driver_version = ".".join(version_info)
                except:  # noqa: E722
                    pass

        if not driver_version:
            driver_version = "UNKOWN_VERSION"
            _logger.warning(
                "Couldn't determine driver version of library: %s.",
                driver_lib.__name__,
            )

        if not base_url.endswith("/"):
            base_url += "/"

        self._http_client.base_url = base_url
        self._http_client.rate_limit = rate_limit
        self._http_client.headers = {
            "User-Agent": f"{self._http_client.driver} {driver_version}; {app_name}",
            "Content-Type": "application/json",
        }
        if api_key:
            self._http_client.headers["Authorization"] = api_key

        for key, value in http_client_kwargs.items():
            setattr(self._http_client, key, value)

    @functools.cache
    def get_runtimes(self) -> List[Runtime]:
        r"""Return a list of available languages."""

        runtime_data = self._http_client.get("runtimes")
        return [Runtime(**runtime) for runtime in runtime_data]

    @aiocache.cached()
    async def get_runtimes_async(self) -> List[Runtime]:
        r"""Return a list of available languages asynchronously."""

        runtime_data = await self._http_client.get_async("runtimes")
        return [Runtime(**runtime) for runtime in runtime_data]

    def get_packages(self) -> List[Package]:
        r"""Returns a list of all possible packages, and their installation status.

        .. warning::
            This method is not available for the public API.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.
        """

        package_data = self._http_client.get("packages")
        return [Package(**package) for package in package_data]

    async def get_packages_async(self) -> List[Package]:
        r"""Returns a list of all possible packages, and their installation status asynchronously.

        .. warning::
            This method is not available for the public API.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.
        """

        package_data = await self._http_client.get_async("packages")
        return [Package(**package) for package in package_data]

    def post_packages(self, language: str, version: str) -> None:
        r"""Install the given package.

        .. warning::
            This method is not available for the public API.

        Parameters
        ----------
        language
            | The name of the programming language.
        version
            | The version of the language.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.

        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """

        json_data: Dict[str, str] = {
            "language": language,
            "version": version,
        }
        self._http_client.post("packages", json_data=json_data)

    async def post_packages_async(self, language: str, version: str) -> None:
        r"""Install the given package asynchronously.

        .. warning::
            This method is not available for the public API.

        Parameters
        ----------
        language
            | The name of the programming language.
        version
            | The version of the language.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.

        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """
        json_data: Dict[str, str] = {
            "language": language,
            "version": version,
        }
        await self._http_client.post_async("packages", json_data=json_data)

    def post_execute(
        self,
        language: str,
        version: str,
        file: File,
        stdin: Optional[str] = None,
        args: Optional[List[str]] = None,
        compile_timeout: Union[float, int] = 10000,
        run_timeout: Union[float, int] = 3000,
        compile_memory_limit: int = -1,
        run_memory_limit: int = -1,
    ) -> ExecutionOutput:
        r"""Runs the given code, using the given runtime and arguments, returning the result.

        Parameters
        ----------
        language
            | The name of the programming language.
        version
            | The version of the language.
        file
            | The file containing the code to be executed.
        stdin
            | Text to pass into stdin of the program.
        args
            | Arguments to pass to the program.
        compile_timeout
            | The maximum allowed time in milliseconds for the run stage to finish before bailing out.
            | Must be a number, less than or equal to the configured maximum timeout. Defaults to maximum.
        run_timeout
            | The maximum allowed time in milliseconds for the compile stage to finish before bailing out.
            | Must be a number, less than or equal to the configured maximum timeout.
        compile_memory_limit
            | The maximum amount of memory the compile stage is allowed to use in bytes.
            | Must be a number, less than or equal to the configured maximum.
            | Defaults to maximum, or -1 (no limit) if none is configured.
        run_memory_limit
            | The maximum amount of memory the run stage is allowed to use in bytes.
            | Must be a number, less than or equal to the configured maximum.
            | Defaults to maximum, or -1 (no limit) if none is configured.

        Returns
        -------
        | Returns the ExecutionOutput data containing various data related to the ran code.

        Raises
        ------
        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """

        json_data: Dict[str, object] = {
            "language": language,
            "version": version,
            "files": [file.model_dump()],
            "compile_timeout": compile_timeout,
            "run_timeout": run_timeout,
            "compile_memory_limit": compile_memory_limit,
            "run_memory_limit": run_memory_limit,
        }
        if stdin:
            json_data["stdin"] = stdin
        if args:
            json_data["args"] = args

        response = self._http_client.post("execute", json_data=json_data)
        return ExecutionOutput(**response)

    async def post_execute_async(
        self,
        language: str,
        version: str,
        file: File,
        stdin: Optional[str] = None,
        args: Optional[List[str]] = None,
        compile_timeout: Union[float, int] = 10000,
        run_timeout: Union[float, int] = 3000,
        compile_memory_limit: int = -1,
        run_memory_limit: int = -1,
    ) -> ExecutionOutput:
        r"""Runs the given code, using the given runtime and arguments, returning the result asynchronously.

        Parameters
        ----------
        language
            | The name of the programming language.
        version
            | The version of the language.
        files
            | The file containing the code to be executed.
        stdin
            | Text to pass into stdin of the program.
        args
            | Arguments to pass to the program.
        compile_timeout
            | The maximum allowed time in milliseconds for the run stage to finish before bailing out.
            | Must be a number, less than or equal to the configured maximum timeout. Defaults to maximum.
        run_timeout
            | The maximum allowed time in milliseconds for the compile stage to finish before bailing out.
            | Must be a number, less than or equal to the configured maximum timeout.
        compile_memory_limit
            | The maximum amount of memory the compile stage is allowed to use in bytes.
            | Must be a number, less than or equal to the configured maximum.
            | Defaults to maximum, or -1 (no limit) if none is configured.
        run_memory_limit
            | The maximum amount of memory the run stage is allowed to use in bytes.
            | Must be a number, less than or equal to the configured maximum.
            | Defaults to maximum, or -1 (no limit) if none is configured.

        Returns
        -------
        Returns the ExecutionOutput data containing various data related to the ran code.

        Raises
        ------
        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """

        json_data: Dict[str, Any] = {
            "language": language,
            "version": version,
            "files": [file.model_dump()],
            "compile_timeout": compile_timeout,
            "run_timeout": run_timeout,
            "compile_memory_limit": compile_memory_limit,
            "run_memory_limit": run_memory_limit,
        }
        if stdin:
            json_data["stdin"] = stdin
        if args:
            json_data["args"] = args

        response = await self._http_client.post_async(
            "execute", json_data=json_data
        )
        return ExecutionOutput(**response)

    def delete_packages(self, language: str, version: str) -> None:
        r"""Uninstall the given package.

        .. warning::
            This method is not available for the public API.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.

        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """

        json_data: Dict[str, str] = {
            "language": language,
            "version": version,
        }
        self._http_client.delete("packages", json_data=json_data)

    async def delete_packages_async(self, language: str, version: str) -> None:
        r"""Uninstall the given package asynchronously.

        .. warning::
            This method is not available for the public API.

        Raises
        ------
        :py:exc:`ultra_piston.errors.NotFoundError`
            | If the endpoint is inaccessible.

        :py:exc:`ultra_piston.errors.BadRequestError`
            | Raised if an invalid version or language was passed.
        """

        json_data: Dict[str, str] = {
            "language": language,
            "version": version,
        }
        await self._http_client.delete_async("packages", json_data=json_data)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Optional


__all__ = (
    "BasePistonError",
    "InternalError",
    "ServerError",
    "MissingDataError",
    "BadRequestError",
    "TooManyRequestsError",
    "InternalServerError",
    "NotFoundError",
    "UnexpectedStatusError",
)


class BasePistonError(Exception):
    r"""The base exception to all ultra-piston errors."""


class InternalError(BaseException):
    r"""Errors raised internally by the wrapper."""


class ServerError(BasePistonError):
    r"""Raised for server responses that return a non-2xx (error) HTTP status code.

    Attributes
    ----------
    endpoint
        | The endpoint URL.
    status_code
        | The status code of the server response.
    """

    def __init__(
        self, *args: Any, endpoint: Optional[str] = None, status_code: int
    ) -> None:
        super().__init__(*args)

        self.endpoint: Optional[str] = endpoint
        self.status_code: int = status_code


class MissingDataError(InternalError):
    r"""Raised when the required data is not set or is missing."""


class BadRequestError(ServerError):
    r"""Raised due to passing invalid arguments.

    Attributes
    ----------
    endpoint
        | The endpoint URL.
    status_code
        | The status code of the server response.
    """

    def __init__(
        self, endpoint: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        STATUS_CODE = 400
        message = message or (
            "Raised due to passing invalid arguments. "
            f"status_code={STATUS_CODE} | endpoint={endpoint}"
        )
        super().__init__(
            message,
            endpoint=endpoint,
            status_code=STATUS_CODE,
        )


class TooManyRequestsError(ServerError):
    r"""Raised due to sending too many requests in a short interval.

    Attributes
    ----------
    endpoint
        | The endpoint URL.
    status_code
        | The status code of the server response.
    """

    def __init__(
        self, endpoint: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        STATUS_CODE = 429
        message = message or (
            "Raised due to sending too many requests in a short interval. "
            f"status_code={STATUS_CODE} | endpoint={endpoint}"
        )
        super().__init__(
            message,
            endpoint=endpoint,
            status_code=STATUS_CODE,
        )


class InternalServerError(ServerError):
    r"""Raised due to an issue with the server.

    Attributes
    ----------
    endpoint
        | The endpoint URL.
    status_code
        | The status code of the server response.
    """

    def __init__(
        self, endpoint: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        STATUS_CODE = 500
        message = message or (
            "Raised due to an issue with the server. "
            f"status_code={STATUS_CODE} | endpoint={endpoint}"
        )

        super().__init__(
            message,
            endpoint=endpoint,
            status_code=STATUS_CODE,
        )


class NotFoundError(ServerError):
    r"""Raised when trying to access an unkown endpoint.

    Attributes
    ----------
    endpoint : Optional[str]
        | The endpoint URL.
    status_code : int
        | The status code of the server response.
    """

    def __init__(
        self, endpoint: Optional[str] = None, message: Optional[str] = None
    ) -> None:
        STATUS_CODE = 404
        message = message or (
            "Tried accessing an unkown endpoint. "
            f"status_code={STATUS_CODE} | endpoint={endpoint}"
        )

        super().__init__(
            message,
            endpoint=endpoint,
            status_code=STATUS_CODE,
        )


class UnexpectedStatusError(ServerError):
    r"""Raised for any unkown response status code (non-2xx).

    Attributes
    ----------
    endpoint : Optional[str]
        | The endpoint URL.
    status_code : int
        | The status code of the server response.
    """

    def __init__(
        self,
        status_code: int,
        endpoint: Optional[str] = None,
        message: Optional[str] = None,
    ) -> None:
        message = message or (
            "Unexpected response code received. "
            f"status_code={status_code} | endpoint={endpoint}"
        )

        super().__init__(
            message,
            endpoint=endpoint,
            status_code=status_code,
        )



from __future__ import annotations

import pytest

from src.ultra_piston import File, PistonClient
from src.ultra_piston.errors import NotFoundError

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
async def test_endpoint_methods() -> None:
    piston = PistonClient()

    result = await piston.get_runtimes_async()
    assert isinstance(result, list)

    with pytest.raises(NotFoundError):
        await piston.get_packages_async()

    with pytest.raises(NotFoundError):
        await piston.post_packages_async("python3", "3.10.0")

    to_be_printed: str = "Hello World"
    code_file = File(content=f"print('{to_be_printed}')")
    executed_output = await piston.post_execute_async(
        "python3", "3.10.0", code_file
    )

    assert executed_output.run.output.strip() == to_be_printed

    with pytest.raises(NotFoundError):
        await piston.delete_packages_async("python3", "3.10.0")



from __future__ import annotations

import pytest

from src.ultra_piston import File, PistonClient
from src.ultra_piston.errors import NotFoundError


def test_endpoint_methods() -> None:
    piston = PistonClient()

    result = piston.get_runtimes()
    assert isinstance(result, list)

    with pytest.raises(NotFoundError):
        piston.get_packages()

    with pytest.raises(NotFoundError):
        piston.post_packages("python3", "3.10.0")

    to_be_printed: str = "Hello World"
    code_file = File(content=f"print('{to_be_printed}')")
    executed_output = piston.post_execute("python3", "3.10.0", code_file)

    assert executed_output.run.output.strip() == to_be_printed

    with pytest.raises(NotFoundError):
        piston.delete_packages("python3", "3.10.0")
